/* bobint.c - bytecode interpreter */
/*
	Copyright (c) 1991, by David Michael Betz
	All rights reserved
*/

#include <setjmp.h>
#include <stdlib.h>
#include "bob.h"
#ifndef BP_RUN
  #include <conio.h>
  int step=0;
#endif
#define isnumtype(x) (((x)->v_type == DT_INTEGER) || ((x)->v_type == DT_FLOAT))
#define iszero(x)	(isnumtype(x) && (x)->v.v_integer == 0)
#define istrue(x)	((x)->v_type != DT_NIL && !iszero(x))

/* global variables */
VALUE *stkbase;		/* the runtime stack */
VALUE *stktop;		/* the top of the stack */
VALUE *sp;		/* the stack pointer */
VALUE *fp;		/* the frame pointer */
int trace=0;		/* variable to control tracing */
#ifndef BP_COMP

/* external variables */
extern DICTIONARY *symbols;
extern jmp_buf error_trap;

/* local variables */
static unsigned char *cbase;	/* the base code address */
static unsigned char *pc;	/* the program counter */
static VALUE *code;		/* the current code vector */

/* forward declarations */
char *typename();

/* execute - execute a bytecode function */
int execute(name)
  char *name;
{
    DICT_ENTRY *sym;

    /* setup an error trap handler */
    if (setjmp(error_trap) != 0)
	return (FALSE);

    /* lookup the symbol */
    if ((sym = findentry(symbols,name)) == NULL)
	return (FALSE);

    /* dispatch on its data type */
    switch (sym->de_value.v_type) {
    case DT_CODE:
	(*sym->de_value.v.v_code)(0);
	break;
    case DT_BYTECODE:
	interpret(sym->de_value.v.v_bytecode);
	break;
    }
    return (TRUE);
}

/* compare helper for interpret */
static long compare()
{
	long result = 0;
	/* register int cnt; */
#ifdef USE_FLOAT
    float tmp;
#endif
		switch (sp[1].v_type)
		{
		case DT_INTEGER:
		    switch (sp[0].v_type)
		    {
		    case DT_INTEGER:
			   result = sp[1].v.v_integer - sp->v.v_integer;
			break;
#ifdef USE_FLOAT
            case DT_FLOAT:
               tmp = (((float) (sp[1].v.v_integer)) - sp->v.v_float);
               if (tmp < 0.0) result = -1;
               else if (tmp > 0.0) result = 1;
            break;
#endif
		    default:
		      badtype(0,DT_INTEGER);
			break;
		    }
		    break;
#ifdef USE_FLOAT
        case DT_FLOAT:
            switch (sp[0].v_type)
            {
			case DT_FLOAT:
			   tmp = sp[1].v.v_float - sp->v.v_float;
			    if (tmp < 0.0) result = -1;
			    else if (tmp > 0.0) result = 1;
			   break;
			case DT_INTEGER:
  			   tmp =sp[1].v.v_float - (float) sp->v.v_integer;
			    if (tmp < 0.0) result = -1;
			    else if (tmp > 0.0) result = 1;
               break;
			default:
		      badtype(0,DT_FLOAT);
			  break;
			}
			break;
#endif
		case DT_STRING:
		   if (sp[0].v_type == DT_STRING)
		   {
			  /* cnt = min(sp[1].v.v_string->s_length,sp[0].v.v_string->s_length); */
			  result = strcmp(sp[1].v.v_string->s_data,sp[0].v.v_string->s_data);
			  /* if (result == 0)
			  {
				  if (cnt < sp[1].v.v_string->s_length) result = 1;
				  if (cnt > sp[1].v.v_string->s_length) result = -1;
			  } */
		   }
		   else if ((sp[0].v_type == DT_NIL))
		      result=1;
		   else
		     badtype(0,DT_STRING);
		   break;
		case DT_NIL:
		   switch(sp[0].v_type)
		   {
			case DT_NIL:
			   result=0;
			   break;
			case DT_OBJECT:
			case DT_CLASS:
			case DT_VECTOR:
			case DT_CODE:
			case DT_BYTECODE:
			case DT_FILE:
			case DT_STRING:
			   result = -1;
			   break;
			default:
			   badtype(0,DT_OBJECT);
			   break;
		   }
		   break;
		default:
		    if (sp[0].v_type == DT_NIL)
		    {
				result = 1;
				break;
			}
			else if(sp[0].v_type == sp[1].v_type)
			{
			   result = (int) (sp[1].v.v_vector) - (int) (sp->v.v_vector);

			}
			else
		       badtype(0,sp[1].v_type);
		    break;
		}
		++sp;
		return result;
}
/* interpret - interpret bytecode instructions */
int interpret(fcn)
  VALUE *fcn;
{
    register int pcoff,n;
    register VALUE *obj;
    long l;
    VALUE *topframe,val;
    STRING *s1,*s2,*sn;

    /* initialize */
    sp = fp = stktop;
    cbase = pc = fcn[1].v.v_string->s_data;
    code = fcn;

    /* make a dummy call frame */
    check(4);
    push_bytecode(code);
    push_integer(0);
    push_integer(0);
    push_integer(0);
    fp = topframe = sp;

    /* execute each instruction */
    for (;;) {
#ifndef BP_RUN
	if (trace || step)
	{
	    decode_instruction(code,pc-code[1].v.v_string->s_data);
		if (step)
		{
		   if (getch() == 27) step = FALSE;
	    }
	}
#endif
	switch (*pc++) {
	case OP_CALL:
		n = *pc++;
		switch (sp[n].v_type) {
		case DT_CODE:
		    (*sp[n].v.v_code)(n);
		    break;
		case DT_BYTECODE:
		    check(3);
		    code = sp[n].v.v_bytecode;
		    push_integer(n);
		    push_integer(stktop - fp);
		    push_integer(pc - cbase);
		    cbase = pc = code[1].v.v_string->s_data;
		    fp = sp;
		    break;
		 case DT_OBJECT:
		    send(n,"OP_CALL",code);
		    break;
		default:
		    error("call to non-procedure, type %s",
			  typename(sp[n].v_type));
		    return;
		}
		break;
	case OP_RETURN:
		if (fp == topframe) return;
		val = *sp;
		sp = fp;
		pcoff = fp[0].v.v_integer;
		n = fp[2].v.v_integer;
		fp = stktop - fp[1].v.v_integer;
		code = fp[fp[2].v.v_integer+3].v.v_bytecode;
		cbase = code[1].v.v_string->s_data;
		pc = cbase + pcoff;
		sp += n + 3;
		*sp = val;
		break;
	case OP_REF:
		*sp = code[*pc++].v.v_var->de_value;
		break;
	case OP_SET:
		code[*pc++].v.v_var->de_value = *sp;
		break;
	case OP_VREF:
		switch (sp[1].v_type) {
		case DT_VECTOR: vectorref(); break;
		case DT_STRING: stringref(); break;
		case DT_OBJECT: send(1,"OP_VREF",code); break;
		default:	badtype(1,DT_VECTOR); break;
		}
		break;
	case OP_VSET:
	     if (sp[2].v_type == DT_OBJECT)
	     {
			 send (2,"OP_VSET",code);
			 break;
		 }
		chktype(1,DT_INTEGER);
		switch (sp[2].v_type) {
		case DT_VECTOR: vectorset(); break;
		case DT_STRING: stringset(); break;
		default:	badtype(1,DT_VECTOR); break;
		}
		break;
	case OP_MREF:
		obj = fp[fp[2].v.v_integer+2].v.v_object;
		*sp = obj[*pc++];
		break;
	case OP_MSET:
		obj = fp[fp[2].v.v_integer+2].v.v_object;
		obj[*pc++] = *sp;
		break;
	case OP_AREF:
		n = *pc++;
		if (n >= fp[2].v.v_integer)
		    error("too few arguments");
		*sp = fp[n+3];
		break;
	case OP_ASET:
		n = *pc++;
		if (n >= fp[2].v.v_integer)
		    error("too few arguments");
		fp[n+3] = *sp;
		break;
	case OP_TREF:
		n = *pc++;
		*sp = fp[-n-1];
		break;
	case OP_TSET:
		n = *pc++;
		fp[-n-1] = *sp;
		break;
	case OP_TSPACE:
		n = *pc++;
		check(n);
		while (--n >= 0) {
		    --sp;
		    set_nil(sp);
		}
		break;
	case OP_BRT:
		if (istrue(sp))
		    pc = cbase + getwoperand();
		else
		    pc += 2;
		break;
	case OP_BRF:
		if (istrue(sp))
		    pc += 2;
		else
		    pc = cbase + getwoperand();
		break;
	case OP_BR:
		pc = cbase + getwoperand();
		break;
	case OP_NIL:
		set_nil(sp);
		break;
	case OP_PUSH:
		check(1);
		push_integer(FALSE);
		break;
	case OP_NOT:
		if (istrue(sp))
		    set_integer(sp,FALSE);
		else
		    set_integer(sp,TRUE);
		break;
	case OP_NEG:
#ifdef USE_FLOAT
        if (sp->v_type == DT_FLOAT)
        {
		   sp->v.v_float = -sp->v.v_float;
		   break;
		}
#endif
		chktype(0,DT_INTEGER);
		sp->v.v_integer = -sp->v.v_integer;
		break;
	case OP_ADD:
		switch (sp[1].v_type) {
		case DT_OBJECT:
		    send(1,"OP_ADD",code);
		    --sp;
//			    sp[1] = sp[0];
	    break;
		case DT_INTEGER:
		    switch (sp[0].v_type) {
		    case DT_INTEGER:
			sp[1].v.v_integer += sp->v.v_integer;
			break;
		    case DT_STRING:
			s2 = sp[0].v.v_string;
			sn = newstring(1 + s2->s_length);
			sn->s_data[0] = sp[1].v.v_integer;
			memcpy(&sn->s_data[1],
			       s2->s_data,
			       s2->s_length);
			set_string(&sp[1],sn);
			break;
#ifdef USE_FLOAT
            case DT_FLOAT:
              set_float(&sp[1],sp[0].v.v_float + (float) sp[1].v.v_integer);
            break;
#endif
		    default:
			  badtype(0,DT_INTEGER);
     		break;
		    }
		    break;
#ifdef USE_FLOAT
        case DT_FLOAT:
            switch (sp[0].v_type)
            {
			case DT_FLOAT:
			   sp[1].v.v_float += sp->v.v_float;
			   break;
			case DT_INTEGER:
  			   sp[1].v.v_float += sp->v.v_integer;
               break;
			default:
		         badtype(0,DT_FLOAT);
			   break;
			}
			break;
#endif
		case DT_STRING:
		    s1 = sp[1].v.v_string;
		    switch (sp[0].v_type) {
		    case DT_INTEGER:
			sn = newstring(s1->s_length + 1);
			memcpy(sn->s_data,
			       s1->s_data,
			       s1->s_length);
			sn->s_data[strlen(s1->s_data)] = sp[0].v.v_integer;
			set_string(&sp[1],sn);
			break;
		    case DT_STRING:
			s2 = sp[0].v.v_string;
			sn = newstring(s1->s_length + s2->s_length);
			strcpy(sn->s_data,s1->s_data);
			strcat(sn->s_data,s2->s_data);
/*			memcpy(sn->s_data,
			       s1->s_data,s1->s_length);
			memcpy(&sn->s_data[s1->s_length],
			       s2->s_data,s2->s_length); */
			set_string(&sp[1],sn);
			break;
		    default:
		      badtype(0,DT_STRING);
			break;
		    }
		    break;
		default:
		    badtype(1,DT_INTEGER);
		    break;
		}
		++sp;
		break;
	case OP_SUB:
		switch (sp[1].v_type) {
		case DT_OBJECT:
		    send(1,"OP_SUB",code);
		    sp--;
		    break;
		case DT_INTEGER:
		    switch (sp[0].v_type) {
		    case DT_INTEGER:
			sp[1].v.v_integer -= sp->v.v_integer;
			break;
#ifdef USE_FLOAT
            case DT_FLOAT:
              set_float(&sp[1],((float) sp[1].v.v_integer) - sp[0].v.v_float);
            break;
#endif
		    default:
		      badtype(0,DT_INTEGER);
			break;
		    }
		    break;
#ifdef USE_FLOAT
        case DT_FLOAT:
            switch (sp[0].v_type)
            {
			case DT_FLOAT:
			   sp[1].v.v_float -= sp->v.v_float;
			   break;
			case DT_INTEGER:
  			   sp[1].v.v_float -= (float) sp->v.v_integer;
               break;
			default:
		      badtype(0,DT_FLOAT);
			  break;
			}
			break;
#endif
		default:
		    badtype(1,DT_INTEGER);
		    break;
		}
/*
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer -= sp->v.v_integer; */
		++sp;
		break;
	case OP_MUL:
		switch (sp[1].v_type) {
		case DT_OBJECT:
		    send(1,"OP_MUL",code);
		    sp--;
		    break;
		case DT_INTEGER:
		    switch (sp[0].v_type) {
		    case DT_INTEGER:
			sp[1].v.v_integer *= sp->v.v_integer;
			break;
#ifdef USE_FLOAT
            case DT_FLOAT:
              set_float(&sp[1],((float) sp[1].v.v_integer) * sp[0].v.v_float);
            break;
#endif
		    default:
		      badtype(0,DT_INTEGER);
			break;
		    }
		    break;
#ifdef USE_FLOAT
        case DT_FLOAT:
            switch (sp[0].v_type)
            {
			case DT_FLOAT:
			   sp[1].v.v_float *= sp->v.v_float;
			   break;
			case DT_INTEGER:
  			   sp[1].v.v_float *= (float) sp->v.v_integer;
               break;
			default:
		      badtype(0,DT_FLOAT);
			  break;
			}
			break;
#endif
		default:
		    badtype(1,DT_INTEGER);
		    break;
		}
		++sp;
		break;
	case OP_DIV:
	     if (sp->v.v_integer == 0) error("divide 0 error");
		switch (sp[1].v_type) {
		case DT_OBJECT:
		    send(1,"OP_DIV",code);
		    --sp;
		    break;
		case DT_INTEGER:
		    switch (sp[0].v_type) {
		    case DT_INTEGER:
			sp[1].v.v_integer /= sp->v.v_integer;
			break;
#ifdef USE_FLOAT
            case DT_FLOAT:
              set_float(&sp[1],((float) sp[1].v.v_integer) / sp[0].v.v_float);
            break;
#endif
		    default:
		      badtype(0,DT_INTEGER);
			break;
		    }
		    break;
#ifdef USE_FLOAT
        case DT_FLOAT:
            switch (sp[0].v_type)
            {
			case DT_FLOAT:
			   sp[1].v.v_float /= sp->v.v_float;
			   break;
			case DT_INTEGER:
  			   sp[1].v.v_float /= (float) sp->v.v_integer;
               break;
			default:
		      badtype(0,DT_FLOAT);
			  break;
			}
			break;
#endif
		default:
		    badtype(1,DT_INTEGER);
		    break;
		}
		++sp;
		break;
	case OP_REM:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_REM",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		if (sp->v.v_integer != 0) {
		    int x=sp->v.v_integer;
		    sp[1].v.v_integer %= x;
		}
		else
		    sp[1].v.v_integer = 0;
		++sp;
		break;
	case OP_INC:
/*	    if (sp[0].v_type == DT_OBJECT)
	    {
			send(0,"OP_INC");
			break;
		} */
		chktype(0,DT_INTEGER);
		++sp->v.v_integer;
		break;
	case OP_DEC:
/*	    if (sp[0].v_type == DT_OBJECT)
	    {
			send(0,"OP_DEC");
			break;
		} */
		chktype(0,DT_INTEGER);
		--sp->v.v_integer;
		break;
	case OP_BAND:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_BAND",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer &= sp->v.v_integer;
		++sp;
		break;
	case OP_BOR:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_BOR",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer |= sp->v.v_integer;
		++sp;
		break;
	case OP_XOR:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_XOR",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer ^= sp->v.v_integer;
		++sp;
		break;
	case OP_BNOT:
		chktype(0,DT_INTEGER);
		sp->v.v_integer = ~sp->v.v_integer;
		break;
	case OP_SHL:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_SHL",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer <<= sp->v.v_integer;
		++sp;
		break;
	case OP_SHR:
	    if (sp[1].v_type == DT_OBJECT)
	    {
			send(1,"OP_SHR",code);
			break;
		}
		chktype(0,DT_INTEGER);
		chktype(1,DT_INTEGER);
		sp[1].v.v_integer >>= sp->v.v_integer;
		++sp;
		break;
	case OP_LT:
	    set_integer(sp,(compare() < 0) ? TRUE : FALSE);
		break;
	case OP_LE:
		set_integer(sp,(compare() <= 0) ? TRUE : FALSE);
		break;
	case OP_EQ:
		set_integer(sp, (compare() == 0) ? TRUE : FALSE);
		break;
	case OP_NE:
		set_integer(sp, (compare() != 0) ? TRUE : FALSE);
		break;
	case OP_GE:
		set_integer(sp, (compare() >= 0) ? TRUE : FALSE);
		break;
	case OP_GT:
		set_integer(sp, (compare() > 0) ? TRUE : FALSE);
		break;
	case OP_LIT:
		*sp = code[*pc++];
		break;
	case OP_SEND:
		n = *pc++;
		chktype(n,DT_OBJECT);
		send(n,0,code);
		break;
	case OP_DUP2:
		check(2);
		sp -= 2;
		*sp = sp[2];
		sp[1] = sp[3];
		break;
	case OP_NEW:
		chktype(0,DT_CLASS);
		set_object(sp,newobject(sp->v.v_class));
		break;
	case OP_DELETE:
		chktype(0,DT_OBJECT);
		free(sp->v.v_object);
		set_nil(sp);
		break;
	case OP_TRON:
#ifndef BP_RUN
        trace = TRUE;
#endif
        break;
	case OP_TROFF:
#ifndef BP_RUN
        trace = FALSE;
#endif
        break;
	case OP_TRSTEP:
#ifndef BP_RUN
		step = TRUE;
#endif
        break;
	default:
		info("Bad opcode %02x",pc[-1]);
		break;
	}
    }
}

/* send - send a message to an object */
send(n,sel,caller)
  int n;
  char* sel;
  VALUE* caller;
{
	int i;
    char selector[TKNSIZE+1],c0;
    char* pselector;
    DICT_ENTRY *de;
    CLASS *bpclass,*pcl;
    bpclass = sp[n].v.v_object[OB_CLASS].v.v_class;
    if (sel == 0)
    {
    	getcstring(selector,sizeof(selector),sp[n-1].v.v_string);
    	sp[n-1] = sp[n];
    	pselector = selector;
	}
	else
	{
	   pselector = sel;
	   sp--;
	   for(i=0;i<=n;i++)
	      sp[i] = sp[i+1];
	   n++;
    }
    do
    {
		de = findentry(bpclass->cl_functions,pselector);
		if (de == 0)
		{
			if ((strncmp(pselector,"BC_",3) == 0) || (strncmp(pselector,"DC_",3) == 0))
			{
				c0 = pselector[0];
				pselector += 3;
				if (caller != 0)
				{
					pcl = caller[2].v.v_class;
					if (pcl != 0)
					{
					     bpclass = pcl;
					     if (c0 == 'D')
					       de = findentry(bpclass->cl_functions,pselector);
				    }
				}
			}
		}
	    if (de != NULL)
	    {
			switch (de->de_value.v_type)
			{
				case DT_CODE:
					(*de->de_value.v.v_code)(n);
					return;
				case DT_BYTECODE:
					check(3);
					code = de->de_value.v.v_bytecode;
					set_bytecode(&sp[n],code);
					push_integer(n);
					push_integer(stktop - fp);
					push_integer(pc - cbase);
					cbase = pc = code[1].v.v_string->s_data;
					fp = sp;
			//		if (strcmp(pselector,"dtor") != 0)
					   return;
					break;
				default:
					error("bad method, name '%s', Type %d",
						  pselector,
						  de->de_value.v_type);
	        }
	    }
    }
    while ((bpclass = bpclass->cl_base) != NULL);
	if ((strcmp(pselector,"dtor") != 0)&&(strcmp(pselector,"BC_dtor") != 0))
         nomethod(selector);
}

/* vectorref - load a vector element */
static vectorref()
{
    VALUE *vect;
    int i;
    vect = sp[1].v.v_vector;
	chktype(0,DT_INTEGER);
    i = sp[0].v.v_integer;
    if (i < 0 || i >= vect[0].v.v_integer)
	error("subscript out of range");
    sp[1] = vect[i+1];
    ++sp;
}

/* vectorset - set a vector element */
static vectorset()
{
    VALUE *vect;
    int i;
    vect = sp[2].v.v_vector;
    i = sp[1].v.v_integer;
    if (i < 0 || i >= vect[0].v.v_integer)
	error("subscript out of range");
    vect[i+1] = sp[2] = *sp;
    sp += 2;
}

/* stringref - load a string element */
static stringref()
{
    STRING *str;
    int i;
    str = sp[1].v.v_string;
	chktype(0,DT_INTEGER);
    i = sp[0].v.v_integer;
    if (i < 0 || i >= str->s_length)
	error("subscript out of range");
    set_integer(&sp[1],str->s_data[i]);
    ++sp;
}

/* stringset - set a string element */
static stringset()
{
    STRING *str;
    int i;
    chktype(0,DT_INTEGER);
    str = sp[2].v.v_string;
    i = sp[1].v.v_integer;
    if (i < 0 || i >= str->s_length)
	error("subscript out of range");
    str->s_data[i] = sp[0].v.v_integer;
    set_integer(&sp[2],str->s_data[i]);
    sp += 2;
}

/* getwoperand - get data word */
static int getwoperand()
{
    int b;
    b = *pc++;
    return ((*pc++ << 8) | b);
}

/* type names */
static char *tnames[] = {
"null","class","object","vector","integer","string","bytecode",
"code","var","FILE", "float"
};

/* typename - get the name of a type */
static char *typename(type)
  int type;
{
    static char buf[20];
    if (type >= _DTMIN && type <= _DTMAX)
	return (tnames[type]);
    sprintf(buf,"(%d)",type);
    return (buf);
}

/* nomethod - report a failure to find a method for a selector */
static nomethod(selector)
  char *selector;
{
    error("no method for name '%s'",selector);
}

/* badtype - report a bad operand type */
badtype(off,type)
  int off,type;
{
    char tn1[20];
    strcpy(tn1,typename(sp[off].v_type));
    info("PC: %04x, offset %d, type %s, expected %s",
	 pc-cbase,off,tn1,typename(type));
    error("invalid argument type");
}


#endif /* ! BP_COMP */

/* stackover - report a stack overflow error */
stackover()
{
    error("stack overflow");
}

